@@ -2428,11 +2428,6 @@ print_tz_as_sql(const char* tz_name, const TIME_ZONE_INFO *sp)
 }
 
 
-#define SAVE_ENGINE(e) \
-  "\"select ENGINE into @" e "_engine" \
-  " from information_schema.TABLES" \
-  " where TABLE_SCHEMA=DATABASE() and TABLE_NAME='" e "'\""
-
 /*
   Print info about leap seconds in time zone as SQL statements
   populating mysql.time_zone_leap_second table.
@@ -2451,11 +2446,12 @@ print_tz_leaps_as_sql(const TIME_ZONE_INFO *sp)
     For all timezones.
   */
   if (!opt_skip_write_binlog)
-      printf(
-        "execute immediate if(@wsrep_cannot_replicate_tz, "
-          SAVE_ENGINE("time_zone_leap_second") ", 'do 0');\n"
-        "execute immediate if(@wsrep_cannot_replicate_tz, "
-	  "'ALTER TABLE time_zone_leap_second ENGINE=InnoDB', 'do 0');\n");
+      printf("\\d |\n"
+        "IF (select count(*) from information_schema.global_variables where\n"
+        "variable_name='wsrep_on' and variable_value='ON') = 1 THEN\n"
+        "ALTER TABLE time_zone_leap_second ENGINE=InnoDB;\n"
+        "END IF|\n"
+        "\\d ;\n");
 
   printf("TRUNCATE TABLE time_zone_leap_second;\n");
 
@@ -2470,10 +2466,12 @@ print_tz_leaps_as_sql(const TIME_ZONE_INFO *sp)
   }
 
   if (!opt_skip_write_binlog)
-      printf(
-        "execute immediate if(@wsrep_cannot_replicate_tz, "
-          "concat('ALTER TABLE time_zone_leap_second ENGINE=', "
-	    "@time_zone_leap_second_engine), 'do 0');\n");
+      printf("\\d |\n"
+        "IF (select count(*) from information_schema.global_variables where\n"
+        "variable_name='wsrep_on' and variable_value='ON') = 1 THEN\n"
+        "ALTER TABLE time_zone_leap_second ENGINE=Aria;\n"
+        "END IF|\n"
+        "\\d ;\n");
 
   printf("ALTER TABLE time_zone_leap_second ORDER BY Transition_time;\n");
 }
@@ -2526,7 +2524,7 @@ scan_tz_dir(char * name_end, uint symlink_recursion_level, uint verbose)
 {
   MY_DIR *cur_dir;
   char *name_end_tmp;
-  uint i;
+  size_t i;
 
   /* Sort directory data, to pass mtr tests on different platforms. */
   if (!(cur_dir= my_dir(fullname, MYF(MY_WANT_STAT|MY_WANT_SORT))))
@@ -2640,33 +2638,24 @@ static struct my_option my_long_options[] =
   {"help", '?', "Display this help and exit.", 0, 0, 0, GET_NO_ARG, NO_ARG,
    0, 0, 0, 0, 0, 0},
 #ifdef DBUG_OFF
-  {"debug", '#', "This is a non-debug version. Catch this and exit.",
+  {"debug", '#', "This is a non-debug version. Catch this and exit",
    0,0, 0, GET_DISABLED, OPT_ARG, 0, 0, 0, 0, 0, 0},
 #else
   {"debug", '#', "Output debug log. Often this is 'd:t:o,filename'.",
    0, 0, 0, GET_STR, OPT_ARG, 0, 0, 0, 0, 0, 0},
 #endif
-  {"leap", 'l', "Print the leap second information from the given time zone file. By convention, when --leap is used the next argument is the timezonefile.",
+  {"leap", 'l', "Print the leap second information from the given time zone file. By convention, when --leap is used the next argument is the timezonefile",
    &opt_leap, &opt_leap, 0, GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
-  {"verbose", 'v', "Write non critical warnings.",
+  {"verbose", 'v', "Write non critical warnings",
    &opt_verbose, &opt_verbose, 0, GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
   {"version", 'V', "Output version information and exit.",
    0, 0, 0, GET_NO_ARG, NO_ARG, 0, 0, 0, 0, 0, 0},
-  {"skip-write-binlog", 'S', "Do not replicate changes to time zone tables to the binary log, or to other nodes in a Galera cluster.",
+  {"skip-write-binlog", 'S', "Do not replicate changes to time zone tables to other nodes in a Galera cluster",
    &opt_skip_write_binlog,&opt_skip_write_binlog, 0, GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
   { 0, 0, 0, 0, 0, 0, GET_NO_ARG, NO_ARG, 0, 0, 0, 0, 0, 0}
 };
 
 
-static char **default_argv;
-
-static void free_allocated_data()
-{
-  free_defaults(default_argv);
-  my_end(0);
-}
-
-
 C_MODE_START
 static my_bool get_one_option(const struct my_option *, const char *,
                               const char *);
@@ -2678,21 +2667,11 @@ static void print_version(void)
 	 MYSQL_SERVER_VERSION,SYSTEM_TYPE,MACHINE_TYPE);
 }
 
-static const char *default_timezone_dir= "/usr/share/zoneinfo/";
-
-
 static void print_usage(void)
 {
-  fprintf(stdout, "Create SQL commands for loading system timezeone data for "
-          "MariaDB\n\n");
-  fprintf(stdout, "Usage:\n");
-  fprintf(stdout, " %s [options] timezonedir\n", my_progname);
-  fprintf(stdout, "or\n");
-  fprintf(stdout, " %s [options] timezonefile timezonename\n", my_progname);
-
-  fprintf(stdout, "\nA typical place for the system timezone directory is "
-          "\"%s\"\n", default_timezone_dir);
-
+  fprintf(stderr, "Usage:\n");
+  fprintf(stderr, " %s [options] timezonedir\n", my_progname);
+  fprintf(stderr, " %s [options] timezonefile timezonename\n", my_progname);
   print_defaults("my",load_default_groups);
   puts("");
   my_print_help(my_long_options);
@@ -2713,39 +2692,19 @@ get_one_option(const struct my_option *opt, const char *argument, const char *)
     print_version();
     puts("");
     print_usage();
-    free_allocated_data();
     exit(0);
   case 'V':
     print_version();
-    free_allocated_data();
     exit(0);
   }
   return 0;
 }
 
 
-static const char *lock_tables=
-  "LOCK TABLES time_zone WRITE,\n"
-  "  time_zone_leap_second WRITE,\n"
-  "  time_zone_name WRITE,\n"
-  "  time_zone_transition WRITE,\n"
-  "  time_zone_transition_type WRITE";
-static const char *trunc_tables_const=
-  "TRUNCATE TABLE time_zone;\n"
-  "TRUNCATE TABLE time_zone_name;\n"
-  "TRUNCATE TABLE time_zone_transition;\n"
-  "TRUNCATE TABLE time_zone_transition_type;\n";
-static const char *wsrep_is_on=
-  "select sum(VARIABLE_NAME='wsrep_on' AND SESSION_VALUE='ON')"
-  " from information_schema.SYSTEM_VARIABLES";
-static const char *wsrep_cannot_replicate_tz=
-  "select sum(VARIABLE_NAME='wsrep_mode' and GLOBAL_VALUE NOT LIKE @replicate_opt)"
-  " from information_schema.SYSTEM_VARIABLES";
-
 int
 main(int argc, char **argv)
 {
-  const char *trunc_tables= "";
+  char **default_argv;
   MY_INIT(argv[0]);
 
   load_defaults_or_exit("my", load_default_groups, &argc, &argv);
@@ -2757,51 +2716,34 @@ main(int argc, char **argv)
   if ((argc != 1 && argc != 2) || (opt_leap && argc != 1))
   {
     print_usage();
-    free_allocated_data();
+    free_defaults(default_argv);
     return 1;
   }
 
-  if (argc == 1 && !opt_leap)
-    trunc_tables= trunc_tables_const;
-
-  printf("set @wsrep_is_on=(%s);\n", wsrep_is_on);
-  printf("SELECT concat('%%', GROUP_CONCAT(OPTION), '%%') INTO @replicate_opt "
-         " FROM"
-         "   (SELECT DISTINCT concat('REPLICATE_', UPPER(ENGINE)) AS OPTION"
-         "    FROM information_schema.TABLES"
-         "    WHERE TABLE_SCHEMA=DATABASE()"
-         "      AND TABLE_NAME IN ('time_zone',"
-         "                         'time_zone_name',"
-         "                         'time_zone_transition',"
-         "                         'time_zone_transition_type',"
-         "                         'time_zone_leap_second')"
-         "      AND ENGINE in ('MyISAM',"
-         "                     'Aria')) AS o"
-         " ORDER BY OPTION DESC;\n");
-  printf("set @wsrep_cannot_replicate_tz=@wsrep_is_on AND (%s);\n", wsrep_cannot_replicate_tz);
   if (opt_skip_write_binlog)
-    /* If turn off session wsrep if we cannot replicate using galera.
-       Disable sql_log_bin as the name implies. */
-    printf("execute immediate if(@wsrep_is_on, 'SET @save_wsrep_on=@@WSREP_ON, WSREP_ON=OFF', 'do 0');\n"
-           "SET @save_sql_log_bin=@@SQL_LOG_BIN;\n"
-           "SET SESSION SQL_LOG_BIN=0;\n"
-           "SET @wsrep_cannot_replicate_tz=0;\n"
-           "%s%s;\n", trunc_tables, lock_tables);
+  {
+    /* If skip_write_binlog is set and wsrep is compiled in we disable
+       sql_log_bin and wsrep_on to avoid Galera replicating below
+       truncate table clauses. This will allow user to set different
+       time zones to nodes in Galera cluster. */
+    printf("set @prep1=if((select count(*) from information_schema.global_variables where variable_name='wsrep_on' and variable_value='ON'), 'SET SESSION SQL_LOG_BIN=?, WSREP_ON=OFF;', 'do ?');\n"
+           "prepare set_wsrep_write_binlog from @prep1;\n"
+           "set @toggle=0; execute set_wsrep_write_binlog using @toggle;\n");
+  }
   else
-    // Alter time zone tables to InnoDB if wsrep_on is enabled
-    // to allow changes to them to replicate with Galera
-    printf(
-      "execute immediate if(@wsrep_cannot_replicate_tz, " SAVE_ENGINE("time_zone") ", 'do 0');\n"
-      "execute immediate if(@wsrep_cannot_replicate_tz, 'ALTER TABLE time_zone ENGINE=InnoDB', 'do 0');\n"
-      "execute immediate if(@wsrep_cannot_replicate_tz, " SAVE_ENGINE("time_zone_name") ", 'do 0');\n"
-      "execute immediate if(@wsrep_cannot_replicate_tz, 'ALTER TABLE time_zone_name ENGINE=InnoDB', 'do 0');\n"
-      "execute immediate if(@wsrep_cannot_replicate_tz, " SAVE_ENGINE("time_zone_transition") ", 'do 0');\n"
-      "execute immediate if(@wsrep_cannot_replicate_tz, 'ALTER TABLE time_zone_transition ENGINE=InnoDB', 'do 0');\n"
-      "execute immediate if(@wsrep_cannot_replicate_tz, " SAVE_ENGINE("time_zone_transition_type") ", 'do 0');\n"
-      "execute immediate if(@wsrep_cannot_replicate_tz, 'ALTER TABLE time_zone_transition_type ENGINE=InnoDB', 'do 0');\n"
-      "%s"
-      "/*M!100602 execute immediate if(@wsrep_cannot_replicate_tz, 'start transaction', '%s')*/;\n"
-      , trunc_tables, lock_tables);
+  {
+      // Alter time zone tables to InnoDB if wsrep_on is enabled
+      // to allow changes to them to replicate with Galera
+      printf("\\d |\n"
+        "IF (select count(*) from information_schema.global_variables where\n"
+        "variable_name='wsrep_on' and variable_value='ON') = 1 THEN\n"
+        "ALTER TABLE time_zone ENGINE=InnoDB;\n"
+        "ALTER TABLE time_zone_name ENGINE=InnoDB;\n"
+        "ALTER TABLE time_zone_transition ENGINE=InnoDB;\n"
+        "ALTER TABLE time_zone_transition_type ENGINE=InnoDB;\n"
+        "END IF|\n"
+        "\\d ;\n");
+  }
 
   if (argc == 1 && !opt_leap)
   {
@@ -2809,6 +2751,12 @@ main(int argc, char **argv)
 
     root_name_end= strmake_buf(fullname, argv[0]);
 
+    printf("TRUNCATE TABLE time_zone;\n");
+    printf("TRUNCATE TABLE time_zone_name;\n");
+    printf("TRUNCATE TABLE time_zone_transition;\n");
+    printf("TRUNCATE TABLE time_zone_transition_type;\n");
+    printf("START TRANSACTION;\n");
+
     if (scan_tz_dir(root_name_end, 0, opt_verbose))
     {
       printf("ROLLBACK;\n");
@@ -2819,15 +2767,11 @@ main(int argc, char **argv)
       return 1;
     }
 
-    printf("UNLOCK TABLES;\n"
-           "COMMIT;\n");
-    printf(
-      "execute immediate if(@wsrep_cannot_replicate_tz, 'do 0',"
-        "'ALTER TABLE time_zone_transition "
-          "ORDER BY Time_zone_id, Transition_time');\n"
-      "execute immediate if(@wsrep_cannot_replicate_tz, 'do 0',"
-        "'ALTER TABLE time_zone_transition_type "
-          "ORDER BY Time_zone_id, Transition_type_id');\n");
+    printf("COMMIT;\n");
+    printf("ALTER TABLE time_zone_transition "
+           "ORDER BY Time_zone_id, Transition_time;\n");
+    printf("ALTER TABLE time_zone_transition_type "
+           "ORDER BY Time_zone_id, Transition_type_id;\n");
   }
   else
   {
@@ -2847,29 +2791,25 @@ main(int argc, char **argv)
       print_tz_leaps_as_sql(&tz_info);
     else
       print_tz_as_sql(argv[1], &tz_info);
-    printf("UNLOCK TABLES;\n"
-           "COMMIT;\n");
+
     free_root(&tz_storage, MYF(0));
   }
 
-  if(opt_skip_write_binlog)
-    printf("SET SESSION SQL_LOG_BIN=@save_sql_log_bin;\n"
-           "execute immediate if(@wsrep_is_on, 'SET SESSION WSREP_ON=@save_wsrep_on', 'do 0');\n");
-  else
-    // Change back to what it was before
-    printf(
-      "execute immediate if(@wsrep_cannot_replicate_tz, "
-        "concat('ALTER TABLE time_zone ENGINE=', @time_zone_engine), 'do 0');\n"
-      "execute immediate if(@wsrep_cannot_replicate_tz, "
-        "concat('ALTER TABLE time_zone_name ENGINE=', @time_zone_name_engine), 'do 0');\n"
-      "execute immediate if(@wsrep_cannot_replicate_tz, "
-        "concat('ALTER TABLE time_zone_transition ENGINE=', "
-	  "@time_zone_transition_engine, ', ORDER BY Time_zone_id, Transition_time'), 'do 0');\n"
-      "execute immediate if(@wsrep_cannot_replicate_tz, "
-        "concat('ALTER TABLE time_zone_transition_type ENGINE=', "
-	  "@time_zone_transition_type_engine, ', ORDER BY Time_zone_id, Transition_type_id'), 'do 0');\n");
+  if(!opt_skip_write_binlog)
+  {
+      // Fall back to Aria
+      printf("\\d |\n"
+        "IF (select count(*) from information_schema.global_variables where\n"
+        "variable_name='wsrep_on' and variable_value='ON') = 1 THEN\n"
+        "ALTER TABLE time_zone ENGINE=Aria;\n"
+        "ALTER TABLE time_zone_name ENGINE=Aria;\n"
+        "ALTER TABLE time_zone_transition ENGINE=Aria;\n"
+        "ALTER TABLE time_zone_transition_type ENGINE=Aria;\n"
+        "END IF|\n"
+        "\\d ;\n");
+  }
 
-  free_allocated_data();
+  free_defaults(default_argv);
   my_end(0);
   return 0;
 }
